/*
 Copyright (C) 2015 Andrey Tsyganov, UlSPU. All rights reserved.

 Contact information
 -------------------
 Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
 Web    :  http://lmm.ulspu.ru
 e-mail :  andrew.tsyganov@gmail.com 
*/

#include "GA.h"

const char max_bias = 5;             // Maximum bias of the mutated gene
const int  len = 13;                 // String length (without extra null character)
const char goal[] = "Hello, world!"; // String to guess

class GAIndiv
{
    friend class MyGA;
private:
    char data_[len + 1];  
};

class MyGA: public GA<int, GAIndiv>
{
public:
    virtual GAIndiv generate() const;
    virtual int     fitness(GAIndiv const& indiv) const;
    virtual GAIndiv crossover(GAIndiv const& parent1, GAIndiv const& parent2) const;
    virtual GAIndiv mutation(GAIndiv const& indiv) const;
    virtual void    print(GAIndiv const& indiv) const;
};

GAIndiv MyGA::generate() const
{
    GAIndiv indiv;

    indiv.data_[len] = '\0';
    for (int i = 0; i < len; i++) 
        indiv.data_[i] = (char) random(32, 128);
    
    return indiv;    
};

int MyGA::fitness(GAIndiv const& indiv) const
{
    int sum = 0;
    for (int i = 0; i < len; i++)
        sum += abs(indiv.data_[i] - goal[i]);

    return sum;
}

GAIndiv MyGA::crossover(GAIndiv const& parent1, GAIndiv const& parent2) const
{
    // Scattered crossover
    GAIndiv child;

    child.data_[len] = '\0';
    for (int i = 0; i < len; i++)
        // Better (first) parent contributes two times more genes
        child.data_[i] = !random(3) ? parent2.data_[i] : parent1.data_[i];

    return child;
};

GAIndiv MyGA::mutation(GAIndiv const& indiv) const
{
    GAIndiv child;

    int pos = random(len);
    
    child.data_[pos] += - max_bias + (char) random(2 * max_bias + 1);
    
    if (child.data_[pos] < 32) 
        child.data_[pos] = 32;

    return child;
};

void MyGA::print(GAIndiv const& indiv) const
{
    Serial.print(indiv.data_);
};



MyGA ga;
GAIndiv indiv;

time_t start_time, end_time;
int seed;

void setup() 
{
  Serial.begin(115200);
  
  ga.setPopSize(15);
  ga.setEliteCount(2);
  ga.setCrossoverFrac(0.6);
  //ga.setStopCond(GEN_LIM);
  //ga.setGenLim(200);
  ga.setStopCond(FITNESS_LIM);
  ga.setFitnessLim(0);
  //ga.setStopCond(TIME_LIM);
  //ga.setTimeLim(4000000); // 4 sec
  //Serial.println("GA Parameters");
  //ga.printParam();
  
  //seed = analogRead(0);
  seed = 0;
}


void loop() 
{
  randomSeed(seed);
  Serial.println();
  Serial.print("Seed: ");
  Serial.println(seed);

  Serial.println("Start GA");
  start_time = micros();
  ga.run();
  end_time   = micros();
  Serial.println("End GA");

  Serial.print("Elapsed time: ");
  Serial.print((float) (end_time - start_time) / 1000000, 6);
  Serial.println(" second(s)");
  Serial.print("Generations: ");
  Serial.println(ga.getGen());
  indiv = ga.getIndiv(0);
  Serial.print("Best solution: ");
  ga.print(indiv);
  Serial.println();
  Serial.print("Fitness: ");
  Serial.println(ga.fitness(indiv));
  
  seed++;
}
