/*
 Copyright (C) 2015 Andrey Tsyganov, UlSPU. All rights reserved.

 Contact information
 -------------------
 Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
 Web    :  http://lmm.ulspu.ru
 e-mail :  andrew.tsyganov@gmail.com 
*/

#include "SA.h"

const int  change_prob_count = 13;   // Change maximum one out of change_prob_count
const char max_bias = 5;             // Maximum bias of symbols in the neighbor
const int  len = 13;                 // String length (without extra null character)
const char goal[] = "Hello, world!"; // String to guess

class SASolution
{
    friend class MySA;
private:
    char data_[len + 1];  
};

class MySA: public SA<int, SASolution>
{
public:
    virtual SASolution generate() const;
    virtual int        cost(SASolution const& solution) const;
    virtual SASolution neighbour(SASolution const& solution) const;
    virtual void       print(SASolution const& solution) const;
};

SASolution MySA::generate() const
{
    SASolution solution;

    solution.data_[len] = '\0';
    for (int i = 0; i < len; i++) 
        solution.data_[i] = (char) random(32, 128);
    
    return solution;    
};

int MySA::cost(SASolution const& solution) const
{
    int sum = 0;
    for (int i = 0; i < len; i++)
        sum += abs(solution.data_[i] - goal[i]);

    return sum;
}

SASolution MySA::neighbour(SASolution const& solution) const
{
    SASolution neighb;

    for (int i = 0; i < len; i++)
        if (!random(change_prob_count))
        {
            neighb.data_[i] += - max_bias + (char) random(2 * max_bias + 1);
            
            if (neighb.data_[i] < 32) 
                neighb.data_[i] = 32;
        }
        else
            neighb.data_[i] = solution.data_[i];

    return neighb;
};

void MySA::print(SASolution const& solution) const
{
    Serial.print(solution.data_);
};



MySA sa;
SASolution solution;

time_t start_time, end_time;
int seed;

void setup() 
{
  Serial.begin(115200);
  
  //sa.setStopCond(NO_ACCEPTANCE);
  sa.setStopCond(COST_LIM);
  sa.setCostLim(0);
  //sa.setCooling(BOLTZ);
  //sa.printParam();
  
  //seed = analogRead(0);
  seed = 0;
}


void loop() 
{
  randomSeed(seed);
  Serial.println();
  Serial.print("Seed: ");
  Serial.println(seed);

  Serial.println("Start SA");
  start_time = micros();
  sa.run();
  end_time   = micros();
  Serial.println("End SA");

  Serial.print("Elapsed time: ");
  Serial.print((float) (end_time - start_time) / 1000000, 6);
  Serial.println(" second(s)");
  solution = sa.getBest();
  Serial.print("Best solution: ");
  sa.print(solution);
  Serial.println();
  Serial.print("Cost: ");
  Serial.println(sa.cost(solution));
  
  seed++;
}
