/*
 Copyright (C) 2015 Andrey Tsyganov, UlSPU. All rights reserved.

 Contact information
 -------------------
 Andrey Tsyganov, Ulyanovsk State Pedagogical University, Laboratory of Mathematical Modeling
 Web    :  http://lmm.ulspu.ru
 e-mail :  andrew.tsyganov@gmail.com 
*/

#include "SHC.h"

const int  change_prob_count = 13;   // Change maximum one out of change_prob_count
const char max_bias = 5;             // Maximum bias of symbols in the neighbor
const int  len = 13;                 // String length (without extra null character)
const char goal[] = "Hello, world!"; // String to guess

class SHCSolution
{
    friend class MySHC;
private:
    char data_[len + 1];  
};

class MySHC: public SHC<int, SHCSolution>
{
public:
    virtual SHCSolution generate() const;
    virtual int        cost(SHCSolution const& solution) const;
    virtual SHCSolution neighbour(SHCSolution const& solution) const;
    virtual void       print(SHCSolution const& solution) const;
};

SHCSolution MySHC::generate() const
{
    SHCSolution solution;

    solution.data_[len] = '\0';
    for (int i = 0; i < len; i++) 
        solution.data_[i] = (char) random(32, 128);
    
    return solution;    
};

int MySHC::cost(SHCSolution const& solution) const
{
    int sum = 0;
    for (int i = 0; i < len; i++)
        sum += abs(solution.data_[i] - goal[i]);

    return sum;
}

SHCSolution MySHC::neighbour(SHCSolution const& solution) const
{
    SHCSolution neighb;

    for (int i = 0; i < len; i++)
        if (!random(change_prob_count))
        {
            neighb.data_[i] += - max_bias + (char) random(2 * max_bias + 1);
            
            if (neighb.data_[i] < 32) 
                neighb.data_[i] = 32;
        }
        else
            neighb.data_[i] = solution.data_[i];

    return neighb;
};

void MySHC::print(SHCSolution const& solution) const
{
    Serial.print(solution.data_);
};



MySHC shc;
SHCSolution solution;

time_t start_time, end_time;
int seed;

void setup() 
{
  Serial.begin(115200);
  
  shc.setStopCond(COST_LIM);
  shc.setCostLim(0);
  //shc.setStopCond(STEP_LIM);
  //shc.setCostLim(1000);
  
  //seed = analogRead(0);
  seed = 0;
}


void loop() 
{
  randomSeed(seed);
  Serial.println();
  Serial.print("Seed: ");
  Serial.println(seed);

  Serial.println("Start SHC");
  start_time = micros();
  shc.run();
  end_time   = micros();
  Serial.println("End SHC");

  Serial.print("Elapsed time: ");
  Serial.print((float) (end_time - start_time) / 1000000, 6);
  Serial.println(" second(s)");
  solution = shc.getBest();
  Serial.print("Best solution: ");
  shc.print(solution);
  Serial.println();
  Serial.print("Cost: ");
  Serial.println(shc.cost(solution));
  
  seed++;
}
